#!/usr/bin/perl

# The DLV system (www.dlvsystem.com) is needed. Specify it's location in $DLV.

$DLV       = 'dlv';
$TRANSFORM = 'br.dl';


# Command line: perl simplify_glsh.pl input.dl
# 
#  or
#
# On unix systems: simplify_glsh.pl input.dl
#
# The execution flag of the file A1.pl has to be set then.

# If there is no file name specified on the command line then the input is read
# from stdin. 

# The input is a file of rules of the following form:
#
#  rule     ::= head '.' | head ':-' body '.' | ':-' body '.'
#  head     ::= literal {'v' literal}
#  body     ::= literal {',' literal}
#  literal  ::= atom | 'not' atom
#  atom     ::= symb ['(' term {',' term}')']
#  term     ::= variable | symb | num
#  variable ::= ('A'-'Z' ) {'a'-'z' | 'A'-'Z' | '0'-'9'}
#  symb     ::= ('a'-'z') {'a'-'z' | 'A'-'Z' | '0'-'9'}
#  num      ::= ('0'-'9') {'0'-'9'}

# The output is the rule r if G-LSH(P,r) holds,
#               no output otherwise.
#
# where P is the input program and r is the first rule.
#
# Please see my master thesis for an explanation of this.

# An intermediate file of the name $TRANSFORM is generated.


sub analyze_rule() {
	local($r, @r, @h, @b, @in, $i, $j, $x);
	$r =  $_[0];
	
	@in = split //, $r;
	for ($b=0, $i=0; $i < @in; $i++) {	
		$b = 1 - $b if ($in[$i] eq '(') || ($in[$i] eq ')');
		$in[$i] = ';' if $in[$i] eq ',' && $b;
	}
	$r = join '', @in;
	$r =~ s/\.//g;
	@r = split /:-/, $r;
	@h = split / v /, @r[0];
	@b = split /\s*,\s*/, @r[1];

	$i = 0;	
	foreach $x (@h) {
		$x =~ s/\s*//g;
		$x =~ s/\)//;
		@x = split /\(/, $x;
		if (@x >= 2) {@x = ($x[0], split /;/, $x[1]);}
		@{$head[$i++]} = @x;
	}
	
	$i = 0;
	$j = 0;
	foreach $x (@b) {
		if (index($x, "not") < 0) {
			$x =~ s/\s*//g;
			$x =~ s/\)//;
			@x = split /\(/, $x;
			@x = ($x[0], split /;/, $x[1]);
			@{$body_pos[$i++]} = @x;
		}
		else {
			$x =~ s/not//;
			$x =~ s/\s*//g;
			$x =~ s/\)//;
			@x = split /\(/, $x;
			@x = ($x[0], split /;/, $x[1]);
			@{$body_neg[$j++]} = @x;
		}		
	}
}

sub get_var() {
	local($a, $x);
	$a = shift @{$_[0]};
	foreach $x (@{$_[0]}) {
		if ($x =~ /([A-Z]+([a-z]|[A-Z]|[0-9])*)/) {
			if ((grep /$x/, @var) == 0) {		
				push @var, $x ;
			}
		}
	}
	unshift @{$_[0]}, $a;
}

sub get_const() {
	local($a, $x);
	$a = shift @{$_[0]};
	foreach $x (@{$_[0]}) {
		if ($x =~ /([0-9]+|[a-z]([a-z]|[A-Z]|[0-9])*)/) {
			if ((grep /$x/, @const) == 0) {		
				push @const, $x ;
			}
		}
	}
	unshift @{$_[0]}, $a;
}

sub test() {
	local($first, @const, @const2, $maxar, $file, $x, $y, $z, $s, $t, $u, $atom, $i, $j, @var);

	$first = $input[$_[0]];
	@const = ();
	$maxar = 0;

	open $file, "> $TRANSFORM";

	splice @head, 0, @head;
	splice @body_pos, 0, @body_pos;
	splice @body_neg, 0, @body_neg;

	#Get the constants and the maximal arity.
	foreach $x (@input) {
		&analyze_rule($x);

		foreach $atom (@head, @body_pos, @body_neg) {
			&get_const($atom);

			if (@head > 0 && @body_pos > 0 && @{$atom}-1 > $maxar) {
				$maxar = @{$atom}-1;
			}
		}

		splice @head, 0, @head;
		splice @body_pos, 0, @body_pos;
		splice @body_neg, 0, @body_neg;
	}

	#Generate the first part.
	if (@const == 0) {
		push @const, "1";
	}

	foreach $x (@const) {
		push @const2, "c2_$x";
		push @const2, "c3_$x";
		push @const2, "c4_$x";
	}

	@const = sort (@const, @const2);

	foreach $x (@const) {
		print $file "p_(" . $x . "). ";
	}

	print $file "\n\n";



	#Generate the second part.
	&analyze_rule($first);
	
	if (@head <= 1) {
		#print $file "b_.\n\n";
		close $file;
		return;
	}

	for ($i = 0; $i < @head; $i++) {
		for ($j = $i + 1; $j < @head; $j++) { 
			@var = ();
			&get_var($head[$i]);
			&get_var($head[$j]);
			$s = join ",", (@{$head[$i]}, map {$_=1} 1..($maxar-@{$head[$i]}+1));
			$t = join ",", (@{$head[$j]}, map {$_=1} 1..($maxar-@{$head[$j]}+1));
			print $file "s($s,$t) :- " . (join ", ", map {$_="p_(".$_.")"} @var) . ".\n";
		}
	}



	#Generate the third part.
	foreach $x (@input) {
		&analyze_rule($x);

		if (@head > 0 && @body_pos > 0) {

			foreach $y (@head) {
				$s = join ",", (@{$y}, map {$_=1} 1..($maxar-@{$y}+1));			
				@var = ();
				&get_var($y);
				@var_ = @var;
				foreach $z (@body_pos) {
					&get_var($z);
					$t = join ",", (@{$z}, map {$_=1} 1..($maxar-@{$z}+1));
					print $file "t_($s,$t) :- " . (join ", ", map {$_="p_(".$_.")"} @var) . ".\n";
					@var = @var_;
				}	
			}	
		}

		splice @head, 0, @head;
		splice @body_pos, 0, @body_pos;
		splice @body_neg, 0, @body_neg;
	}

	print $file "\n";
	$s = join ",", (map {$_="X_".$_} 1..$maxar+1);
	$t = join ",", (map {$_="Y_".$_} 1..$maxar+1);
	$u = join ",", (map {$_="Z_".$_} 1..$maxar+1);
	print $file "t_($s,$u) :- t_($s,$t), t_($t,$u).\n\n";
	print $file "b_ :- t_($s,$t), t_($t,$s), s($s,$t).\n\n";
	print $file "b_ :- s($s,$s).\n\n";
	close $file;



	#One oracle query.
	if (index(`$DLV $TRANSFORM`, "b_") >= 0) {
		#print "FALSE\n";
	} else {
		#print "TRUE\n";
		print $first . ".\n\n";
	}
}


#Main program.
#open $file, "< input.dl";
#@input = <$file>;
#close $file;
@input = <>;
$input = join / /, @input;
$input =~ s/\n//g;
@input = split /\./, $input;

for ($i = 0; $i < @input; $i++) {
	&test($i);
}

