#!/usr/bin/perl

# The DLV system (www.dlvsystem.com) is needed. Specify it's location in $DLV.

$DLV       = 'dlv';
$TRANSFORM = 'bcq.dl';
$DETAILS   = 1;


# Command line: perl simplify_gsubs.pl input.dl
# 
#  or
#
# On unix systems: simplify_gsubs.pl input.dl
#
# The execution flag of the file simplify_gsubs.pl has to be set then.

# If there is no file name specified on the command line then the input is read
# from stdin. 

# The input is a file of rules of the following form:
#
#  rule     ::= head '.' | head ':-' body '.' | ':-' body '.'
#  head     ::= literal {'v' literal}
#  body     ::= literal {',' literal}
#  literal  ::= atom | 'not' atom
#  atom     ::= symb ['(' term {',' term}')']
#  term     ::= variable | symb | num
#  variable ::= ('A'-'Z' ) {'a'-'z' | 'A'-'Z' | '0'-'9'}
#  symb     ::= ('a'-'z') {'a'-'z' | 'A'-'Z' | '0'-'9'}
#  num      ::= ('0'-'9') {'0'-'9'}

# All other rules then the first two in the input are ignored.

# The output are the rules r, s if SUBS(r,s) holds,
#                no output otherwise.
#
# where r is the first rule and s the second. The order is important. Roughly speaking,
# SUBS(r,s) holds if s subsumes r.
#
# Please see my master thesis for an explanation of this.

# An intermediate file of the name $TRANSFORM is generated.

# The Boolean option $DETAILS changes the output format.

sub analyze_rule() {
	local($r, @r, @h, @b, @in, $i, $j, $x);

	$r =  $_[0];
	@in = split //, $r;
	for ($b=0, $i=0; $i < @in; $i++) {	
		$b = 1 - $b if ($in[$i] eq '(') || ($in[$i] eq ')');
		$in[$i] = ';' if $in[$i] eq ',' && $b;
	}
	$r = join '', @in;
	$r =~ s/\.//g;
	@r = split /:-/, $r;
	@h = split / v /, @r[0];
	@b = split /\s*,\s*/, @r[1];

	$i = 0;	
	foreach $x (@h) {
		$x =~ s/\s*//g;
		$x =~ s/\)//;
		@x = split /\(/, $x;
		if (@x >= 2) {@x = ($x[0], split /;/, $x[1]);}
		@{$head[$i++]} = @x;
		#print "Head:@x\n";
	}
	
	$i = 0;
	$j = 0;
	foreach $x (@b) {
		if (index($x, "not") < 0) {
			$x =~ s/\s*//g;
			$x =~ s/\)//;
			@x = split /\(/, $x;
			@x = ($x[0], split /;/, $x[1]);
			@{$body_pos[$i++]} = @x;
			#print "Body, positive: @x\n";
		}
		else {
			$x =~ s/not//;
			$x =~ s/\s*//g;
			$x =~ s/\)//;
			@x = split /\(/, $x;
			@x = ($x[0], split /;/, $x[1]);
			@{$body_neg[$j++]} = @x;
			#print "Body, negative: @x\n";
		}		
	}
}

sub test() {
	local($in0, $in1, $s, $x, $y, $z, $file);

	$in0 =  $_[0];
	$in1 =  $_[1];

	splice @head, 0, @head;
	splice @body_pos, 0, @body_pos;
	splice @body_neg, 0, @body_neg;
	&analyze_rule($input[$in1]);

	$s = "b__ :- ";
	foreach $x (@head) {
		$s .= (shift @{$x}) . "(";

		foreach $y (@{$x}) {
			$s .= $y . ",";
		}
		chop $s if (@{$x} >= 1);
		$s .= "), ";
	}

	foreach $x (@body_pos) {
		$s .= (shift @{$x}) . "_p(";

		foreach $y (@{$x}) {
			$s .= $y . ",";		
		}
		chop $s if (@{$x} >= 1);
		$s .= "), ";
	}

	foreach $x (@body_neg) {
		$s .= (shift @{$x}) . "_n(";

		foreach $y (@{$x}) {
			$s .= $y . ",";		
		}
		chop $s if (@{$x} >= 1);
		$s .= "), ";
	}
	chop $s;
	chop $s;
	$s .= ".\n\n";
	$s =~ s/\(\)//g;

	#print "$s\n";

	open $file, "> $TRANSFORM";
	print $file $s;

	splice @head, 0, @head;
	splice @body_pos, 0, @body_pos;
	splice @body_neg, 0, @body_neg;
	&analyze_rule($input[$in0]);

	$s = "";
	foreach $x (@head, @body_neg) {
		$z = (shift @{$x});
		$s .= $z . "(";

		foreach $y (@{$x}) {
			$s .= "c_" . $y . ",";
		}
		chop $s if (@{$x} >= 1);
		$s .= ").\n";
		unshift @{$x}, $z;
	}

	foreach $x (@body_pos) {
		$s .= (shift @{$x}) . "_p(";

		foreach $y (@{$x}) {
			$s .= "c_" . $y . ",";		
		}
		chop $s if (@{$x} >= 1);
		$s .= ").\n";
	}

	foreach $x (@body_neg) {
		$s .= (shift @{$x}) . "_n(";

		foreach $y (@{$x}) {
			$s .= "c_" . $y . ",";		
		}
		chop $s if (@{$x} >= 1);
		$s .= ").\n";
	}
	$s =~ s/\(\)//g;

	#print "$s\n";

	print $file $s;
	close $file;

	if (index(`$DLV $TRANSFORM`, "b__") >= 0) {
		#print "TRUE\n";
		if ($DETAILS) {
			print $input[$in0] . ".\n[subsumbed by " . $input[$in1] . ".]\n\n";
		}
		else {
			print $input[$in0] . ".\n" . $input[$in1] . ".\n\n";
		}
	}
	else {
		#print "FALSE\n";
	}
	`rm $TRANSFORM`;
}



#Main program.
#open $file, "< input.dl";
#@input = <$file>;
#close $file;
@input = <>;
$input = join / /, @input;
$input =~ s/\n//g;
@input = split /\./, $input;

for ($i = 0; $i < @input; $i++) {
	for ($j = 0; $j < @input; $j++) {
		if ($i != $j) {
			&test($i, $j);
		}
	}
}



















